# Copyright (C) 2019 Sergej Alikov <sergej.alikov@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import base64

from flask import (
    Flask, render_template, request, make_response, redirect, url_for, session, abort)
from Crypto import Random
from werkzeug.middleware.proxy_fix import ProxyFix

from ephemeral.data import (
    get_message, add_message, MessageNotFoundError, IncorrectPinError)

application = Flask(__name__)
application.wsgi_app = ProxyFix(
    application.wsgi_app, x_for=1, x_host=1, x_proto=1, x_port=1)
application.secret_key = os.environ.get('EPHEMERAL_SECRET_KEY', None)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@application.before_request
def csrf_protect():
    if request.method == 'GET':
        if '_csrf_token' not in session:
            random_token = Random.new().read(32)
            session['_csrf_token'] = base64.b64encode(random_token).decode('latin-1')

    elif request.method == "POST":
        token = session.get('_csrf_token', None)
        if not token or token != request.form.get('_csrf_token'):
            abort(403)


@application.route('/')
def root():
    return render_template('root.html')


@application.route('/show/<msg_id>', methods=['POST', 'GET'])
def show(msg_id):
    if request.method == 'POST':
        try:
            body = get_message(msg_id, request.form['pin'])
        except MessageNotFoundError:
            return render_template('not_found.html', msg_id=msg_id), 404
        except IncorrectPinError:
            pass
        else:
            logger.info('SHOW {}'.format(msg_id))
            response = make_response(render_template('show.html', body=body))

            response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
            response.headers['Pragma'] = 'no-cache'
            response.headers['Expires'] = '0'

            return response

    return render_template('pin.html', csrf_token=session.get('_csrf_token', None))


@application.route('/link/<msg_id>')
def link(msg_id):
    return render_template('link.html', url='{}'.format(
        url_for('show', _external=True, msg_id=msg_id)))


@application.route('/add', methods=['POST', 'GET'])
def add():
    if request.method == 'POST':
        msg_id = add_message(request.form['body'], request.form['pin'])
        return redirect(url_for('link', msg_id=msg_id))
    else:
        return render_template('add.html', csrf_token=session.get('_csrf_token', None))



# Copyright (C) 2019 Sergej Alikov <sergej.alikov@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import uuid
import json
import base64

from Crypto.Hash import SHA256
from Crypto.Cipher import AES
from Crypto import Random
import redis

REDIS_HOST = os.environ.get('EPHEMERAL_REDIS_HOST', '127.0.0.1')
REDIS_PORT = int(os.environ.get('EPHEMERAL_REDIS_PORT', '6379'))
REDIS_DB = int(os.environ.get('EPHEMERAL_REDIS_DB', '0'))

store = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB)


class IncorrectPinError(Exception):
    pass


class MessageNotFoundError(Exception):
    pass


def pad(s):
    pad_by = AES.block_size - len(s) % AES.block_size
    return s + bytes([pad_by]) * pad_by


def unpad(s):
    return s[0:-s[-1]]


def sha256(b):
    h = SHA256.new()
    h.update(b)

    return h.digest()


def b64decode(s):
    return base64.b64decode(s.encode('latin-1'))


def b64encode(b):
    return base64.b64encode(b).decode('latin-1')


def get_message(msg_id, pin):

    try:
        msg = store[msg_id]
    except KeyError:
        raise MessageNotFoundError()

    msg = json.loads(msg)

    iv = b64decode(msg['iv'])
    key = sha256(pin.encode('utf-8'))

    c = AES.new(key, AES.MODE_CBC, iv)

    padded_utf8_body = c.decrypt(b64decode(msg['body']))

    if b64decode(msg['hash']) != sha256(padded_utf8_body):
        raise IncorrectPinError()

    # Delete only when PIN has been confirmed to be correct
    del store[msg_id]

    return unpad(padded_utf8_body).decode('utf-8')


def add_message(body, pin):
    key = sha256(pin.encode('utf-8'))

    iv = Random.new().read(AES.block_size)

    c = AES.new(key, AES.MODE_CBC, iv)

    msg_id = str(uuid.uuid4())

    msg = {}
    msg['iv'] = b64encode(iv)

    padded_utf8_body = pad(body.encode('utf-8'))

    msg['hash'] = b64encode(sha256(padded_utf8_body))
    msg['body'] = b64encode(c.encrypt(padded_utf8_body))

    store[msg_id] = json.dumps(msg)

    return msg_id



#!/usr/bin/env python3

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'VERSION')) as f:
    version = f.read().strip()


setup(
    name='ephemeral',

    version=version,

    description='Self-destructing messages',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/node13h/ephemeral',

    author='Sergej Alikov',
    author_email='sergej.alikov@gmail.com',

    license='AGPL-3.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3',
    ],

    packages=find_packages(exclude=['tests']),
    include_package_data=True,

    # See https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'Flask~=1.1',
        'gunicorn~=19.9',
        'pycrypto~=2.6',
        'redis~=3.2'
    ],

    scripts=['ephemeral.sh'],

    zip_safe=False,
)
